//****************************************************************************
//**
//**    CON_MAIN.CPP
//**    Command Console
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#include "sys_defs.h"
#include "con_main.h"
//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define CON_MAXPRINTLINES   64
#define CON_MAXLINELEN      256
#define CON_MAXARGS         32
//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static con_var_t* con_varList;
static con_func_t* con_funcList;
static con_func_t* con_varHandlerList;
static char con_printLines[CON_MAXPRINTLINES][CON_MAXLINELEN];
static char* con_printLinePtr;
static dword con_printLineIndex;
static void (*con_printCallback)(char*);

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
CON_VARHANDLER(float)
{
    float *val = (float *)argvar->valuePtr;
	if (argc < 2)
	{
		CON_Printf(" float %s == %.4f\n", argvar->name, *val);
		return;
	}
	if (!strcmp(argv[1], "="))
		*val = atof(argv[2]);
	else
	if (!strcmp(argv[1], "+="))
		*val += atof(argv[2]);
	else
	if (!strcmp(argv[1], "-="))
		*val -= atof(argv[2]);
	else
	if (!strcmp(argv[1], "*="))
		*val *= atof(argv[2]);
	else
	if (!strcmp(argv[1], "/="))
		*val /= atof(argv[2]);
	else
		*val = atof(argv[1]);
}

CON_VARHANDLER(int)
{
	int *val = (int *)argvar->valuePtr;
	if (argc < 2)
	{
		CON_Printf(" int %s == %d\n", argvar->name, *val);
		return;
	}
	if (!strcmp(argv[1], "="))
		*val = atoi(argv[2]);
	else
	if (!strcmp(argv[1], "+="))
		*val += atoi(argv[2]);
	else
	if (!strcmp(argv[1], "-="))
		*val -= atoi(argv[2]);
	else
	if (!strcmp(argv[1], "*="))
		*val *= atoi(argv[2]);
	else
	if (!strcmp(argv[1], "/="))
		*val /= atoi(argv[2]);
	else
	if (!strcmp(argv[1], "&="))
		*val &= atoi(argv[2]);
	else
	if (!strcmp(argv[1], "|="))
		*val |= atoi(argv[2]);
	else
	if (!strcmp(argv[1], "^="))
		*val ^= atoi(argv[2]);
	else
	if (!strcmp(argv[1], "%="))
		*val %= atoi(argv[2]);
	else
	if (!stricmp(argv[1], "on"))
		*val = 1;
	else
	if (!stricmp(argv[1], "off"))
		*val = 0;
	else
	if (!stricmp(argv[1], "tog"))
	{
		if (*val)
			*val = 0;
		else
			*val = 1;
	}
	else
		*val = atoi(argv[1]);
}

CON_VARHANDLER(bbool)
{
	int *val = (int *)argvar->valuePtr;
	if (argc < 2)
	{
		if (*val)
			CON_Printf(" bbool %s is on\n", argvar->name, *val);
		else
			CON_Printf(" bbool %s is off\n", argvar->name, *val);
		return;
	}
	if (!strcmp(argv[1], "="))
		*val = atoi(argv[2]);
	else
	if (!strcmp(argv[1], "^="))
		*val ^= atoi(argv[2]);
	else
	if (!stricmp(argv[1], "on"))
		*val = 1;
	else
	if (!stricmp(argv[1], "off"))
		*val = 0;
	else
	if (!stricmp(argv[1], "tog"))
	{
		if (*val)
			*val = 0;
		else
			*val = 1;
	}
	else
		*val = atoi(argv[1]);

	// auto clamp to 0 or 1
	if (*val)
		*val = 1;
}

CON_FUNC(Echo, 0)
{
	static char tbuffer[CON_MAXLINELEN];
	if (argc < 2)
    {
		CON_Printf("\n", tbuffer);
        return;
    }
	tbuffer[0] = 0;
	for (int i=1;i<argc-1;i++)
	{
		strcat(tbuffer, argv[i]);
		strcat(tbuffer, " ");
	}
	strcat(tbuffer, argv[i]);
	CON_Printf("%s\n", tbuffer);
}

CON_FUNC(Help, 0)
{
	CON_Printf("Help: Executing \"listcmds\"\n");
	CON_Execute(false, "listcmds");
}

CON_FUNC(Exec, 0)
{
    if (argc < 2)
    {
        CON_Printf("[?] EXEC filename.cfg\n");
        return;
    }
    if (!CON_ExecuteFile(argv[1]))
        CON_Printf("EXEC: Unable to execute %s\n", argv[1]);
}

CON_FUNC(Listvars, 0)
{
	con_var_t* v;

	for (v=con_varList;v;v=v->next)
	{
		if ((argc < 2) || (!strnicmp(v->name, argv[1], strlen(argv[1]))))
			CON_Printf("  %s (%s)\n", v->name, v->varTypeName);
	}
}

CON_FUNC(Listcmds, 0)
{
	con_func_t* f;
	for (f=con_funcList;f;f=f->next)
	{
		if ((argc < 2) || (!strnicmp(f->name, argv[1], strlen(argv[1]))))
		{
			CON_Printf("  %s\n", f->name);
		}
	}
}

CON_FUNC(Listtypes, 0)
{
	con_func_t* f;
	for (f=con_varHandlerList;f;f=f->next)
	{
		if ((argc < 2) || (!strnicmp(f->name, argv[1], strlen(argv[1]))))
		{
			CON_Printf("  %s\n", f->name);
		}
	}
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// initialization/shutdown

void CON_Init()
{
	int i;
	char execbuf[256];

    con_printLinePtr = con_printLines[0];
    
    CON_Printf(SYS_APPTITLE " console initialized.\n");

	// execute any command line parameter commands first
    execbuf[0] = 0;
	for (i=1;i<SYS_Argc();i++)
	{
		if (SYS_Argv(i)[0] == '+')
		{
			// console operation, exec any current command and start a new one
            if (execbuf[0])
				CON_Execute(false, execbuf);
			strcpy(execbuf, &SYS_Argv(i)[1]);
		}
		else
		if (SYS_Argv(i)[0] == '-')
		{
			// command line option, exec any current command
            if (execbuf[0])
				CON_Execute(false, execbuf);
			execbuf[0] = 0;
		}
		else
		{
			if (execbuf[0])
			{
				// if a command is being built, tack on this argument
                strcat(execbuf, " ");
				strcat(execbuf, SYS_Argv(i));
			}
		}
	}
	// done; if a command is pending, execute it
    if (execbuf[0])
		CON_Execute(false, execbuf);

    // try and execute the app's initial config file as well if it exists
    CON_Execute(false, "exec " SYS_APPTITLE ".cfg");
}

void CON_Shutdown()
{
    // empty for now
}

// var/func operations

void CON_RegisterVar(con_var_t* var)
{
    var->next = con_varList;
    con_varList = var;
}

void CON_RegisterFunc(con_func_t* func)
{
    func->next = con_funcList;
    con_funcList = func;
}

void CON_RegisterVarHandler(con_func_t* varHandler)
{
    varHandler->next = con_varHandlerList;
    con_varHandlerList = varHandler;
}

con_func_t* CON_GetHandlerForVarType(char* varTypeName)
{
    con_func_t* handler;

    if (!varTypeName)
        return(NULL);
    for (handler = con_varHandlerList; handler; handler = handler->next)
    {
        if (!stricmp(handler->name, varTypeName))
            return(handler);
    }
    return(NULL);
}

// command execution

bbool CON_Execute(bbool restricted, char* fmt, ... )
{
	char cmdBuffer[CON_MAXLINELEN];
	char pendingBuffer[CON_MAXLINELEN];
	char *cmdArgv[CON_MAXARGS];
	int cmdArgc;
	char *ptr;
	enum
	{
		STATE_NORMAL,
		STATE_QUOTE
    } state = STATE_NORMAL;
	con_var_t *v, *vlast=NULL;
    con_func_t *f, *flast=NULL;
    va_list args;
    bbool inWhiteSpace = 1;
	bbool found = 0;
	
    va_start(args, fmt);
    vsprintf(cmdBuffer, fmt, args);
    va_end(args);
	
    cmdArgc = 0;
	pendingBuffer[0] = 0;
	for (ptr=cmdBuffer;*ptr;ptr++)
	{
		if (state == STATE_NORMAL)
		{
			if (*ptr == ';')
			{
				// semicolon delimits a command, log it to the pending command buffer
                *ptr = 0;
				strcpy(pendingBuffer, ptr+1);
				ptr--; // so the zero will catch
			}
			else
			if ((*ptr == '/') && (*(ptr+1) == '/'))
			{
				// starting a comment, line's over
                *ptr = 0;
				ptr--; // so the zero will catch
			}
			else
			if (*ptr == '`')
			{
				// replace tilde-key apostrophes with literal quotes
                // it'd be better to use \" directly, but this'll work for now
                *ptr = '\"';
			}
			else
			if (*ptr == ' ')
			{
				// whitespace delimits an argument change impending
                *ptr = 0;
				inWhiteSpace = 1;
			}
			else
			if (inWhiteSpace)
			{
				// we were in whitespace but now we're not
				if (cmdArgc >= CON_MAXARGS)
				{
					// can't take any more arguments
                    *ptr = 0;
					ptr--;
				}
				else
				{
                    // start a new argument
                    cmdArgv[cmdArgc] = ptr;
					cmdArgc++;
					inWhiteSpace = 0;
					if (*ptr == '\"')
					{
						// if the argument starts with quotations, change state
                        state = STATE_QUOTE;
						cmdArgv[cmdArgc-1] = ptr+1; // don't include the quotation character
					}
				}
			}
		}
		else
		if (state == STATE_QUOTE)
		{
			if (*ptr == '\"')
			{
				// quotation state is over
                *ptr = 0;
				state = STATE_NORMAL;
				inWhiteSpace = 1;
			}
			else
			if (*ptr == '`')
			{
				// replace tilde-key apostrophes with literal quotes
                *ptr = '\"';
			}
		}
	}
	// if there's nothing to execute, exit out
    if (!cmdArgc)
		return(0);

	// check convars first
	for (v = con_varList; v; vlast=v, v = v->next)
	{
		if (!stricmp(v->name, cmdArgv[0]))
		{
			// found a match, check restricted
            if ((restricted) && (v->flags & CONVARF_RESTRICTED))
            {
                CON_Printf("Access to %s is restricted\n", v->name);
                found = 1;
                break;
            }
            // see if we have a handler
            if (!v->handler)
			{
				// we don't, maybe it just hasn't linked to it yet
                if (!(v->handler = CON_GetHandlerForVarType((char*)v->varTypeName)))
				{
					// nope, it's really not there
                    CON_Printf("Unknown data type %s\n", v->varTypeName);
					found = 1;
					break;
				}
			}
			// call type handler
			v->handler->func(v, cmdArgc, cmdArgv);
			// move variable to top of list for quicker access next time
            if (vlast)
			{
				vlast->next = v->next;
				v->next = con_varList;
				con_varList = v;
			}
			found = 1;
			break;
		}
	}
	// now check confuncs
	if (!found)
    {
        for (f = con_funcList; f; flast=f, f = f->next)
	    {
		    if (!stricmp(f->name, cmdArgv[0]))
		    {			    
			    // found a match, check restricted
                if ((restricted) && (f->flags & CONFUNCF_RESTRICTED))
                {
                    CON_Printf("Access to %s is restricted\n", f->name);
                    found = 1;
                    break;
                }
                // call the function
			    f->func(NULL, cmdArgc, cmdArgv);
                // move function to top of list for quicker access next time
                if (flast)
			    {
				    flast->next = f->next;
				    f->next = con_funcList;
				    con_funcList = f;
			    }
			    found = 1;
			    break;
		    }
	    }
    }
	if (!found)
		CON_Printf("Unknown variable/cmd: %s\n", cmdArgv[0]);

	// if a command is pending because of a semicolon, execute it
    if (pendingBuffer[0])
		return(CON_Execute(restricted, pendingBuffer));

	return(1);
}

bbool CON_ExecuteFile(char* fileName)
{
	static char execbuf[256];
	static char filebuf[256];
	char *ptr, *cfgbuf;
	int len, elen;
    dword fp;

	strcpy(filebuf, fileName);
	M_SuggestFileExtension(filebuf, "cfg");
	fp = XF_Open(filebuf);
	if (!fp)
		return(0);
	CON_Printf("Execing \"%s\"\n", filebuf);
	len = XF_Size(fp);
	cfgbuf = SYS_MALLOC(char, len+1);
    cfgbuf[len] = 0;
	XF_Read(cfgbuf, len, fp);
	XF_Close(fp);

	// the following code isn't the speediest road runner in the desert, but who cares
    elen = 0;
    execbuf[0] = 0;
	for (ptr=cfgbuf;*ptr;ptr++)
	{
		// NOTE: these 0x0D/0x0A checks are somewhat platform-dependent
        switch(*ptr)
        {
        case 0x0D: break; // skip carriage return
        case 0x0A:
            // execute current command on line feed
            if (!execbuf[0])
                break;
            CON_Execute(false, execbuf);
            elen = 0;
            execbuf[0] = 0;
            break;
        default:
            // any other character gets tacked on
            execbuf[elen++] = *ptr;
            execbuf[elen] = 0;
            break;
        }
	}
	SYS_FREE(cfgbuf);
    return(1);
}

char* CON_MatchCommand(char* nameStart, dword nameCmpLen, dword matchSkipCount)
{
	dword oldMatches = matchSkipCount;
	char *lastMatch=NULL;
	con_var_t *v;
    con_func_t *f;

    while (1)
	{
		// check all variables first
        for (v = con_varList; v; v = v->next)
		{
			if (!strnicmp(v->name, nameStart, nameCmpLen))
			{
				// found a match, log it and return or keep going
				lastMatch = (char*)v->name;
				if (!matchSkipCount)
					return(lastMatch);
				else
					matchSkipCount--;
			}
		}
		// then check functions
        for (f = con_funcList; f; f = f->next)
		{
			if (!strnicmp(f->name, nameStart, nameCmpLen))
			{
				// found a match, log it and return or keep going
                lastMatch = (char*)f->name;
				if (!matchSkipCount)
					return(lastMatch);
				else
					matchSkipCount--;
			}
		}
		// we haven't found a single match yet, so it's all over
        if (matchSkipCount == oldMatches)
			return(NULL);
	}
}

// printing
void CON_Printf(char* fmt, ... )
{
    char* ptr;
    char buffer[2048];
    va_list args;

    va_start(args, fmt);
    vsprintf(buffer, fmt, args);
    va_end(args);
    
    if (con_printCallback)
        con_printCallback(buffer);

    // write the string to the printlines, processing '\n's as needed
    for (ptr=buffer; *ptr; ptr++)
    {
        if (*ptr == '\n')
        {
            con_printLineIndex++;
            con_printLineIndex %= CON_MAXPRINTLINES;
            con_printLinePtr = con_printLines[con_printLineIndex];
        }
        else
        {
            *con_printLinePtr++ = *ptr;
        }            
    }
}

dword CON_GetNumPrintLines()
{
    return(CON_MAXPRINTLINES);
}

char* CON_GetPrintLine(dword index)
{
    return(con_printLines[(con_printLineIndex - index) % CON_MAXPRINTLINES]);
}

void CON_SetPrintCallback(void (*callback)(char*))
{
    con_printCallback = callback;
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================
con_func_t::con_func_t(const char* inName, con_func_bodyFunc_t inFunc,
                       dword inFlags, bbool isVarHandler)
{
    name = inName;
    func = inFunc;
    flags = inFlags;
    next = NULL;
    if (isVarHandler)
        CON_RegisterVarHandler(this);
    else
        CON_RegisterFunc(this);
}

con_func_t::~con_func_t()
{
}

con_var_t::con_var_t(const char* inName, void* inValuePtr,
                     const char* inVarTypeName, dword inFlags)
{
    name = inName;
    varTypeName = inVarTypeName;
    valuePtr = inValuePtr;
    flags = inFlags;
    handler = CON_GetHandlerForVarType((char*)varTypeName); // handler may be NULL if handler hasn't registered yet; that's ok
    next = NULL;
    CON_RegisterVar(this);
}

con_var_t::~con_var_t()
{
}

//****************************************************************************
//**
//**    END MODULE CON_MAIN.CPP
//**
//****************************************************************************

