//****************************************************************************
//**
//**    VID_MAIN.CPP
//**    Video Interface
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32 // for BMP info, not for vid_win
#include "sys_defs.h"
#include "vid_main.h"
#include "vid_win.h"

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static void* BmpResourceLoad(char* fileName, void* userData)
{
	dword fp;
	byte* fpBuf;
    dword result;
    BITMAPFILEHEADER* fhdr;
    BITMAPINFO* info;
    void* totalData = NULL;
    void* imageData = NULL;
    dword i;

	if (!(fp = XF_Open(fileName)))
        return(NULL);
    fpBuf = SYS_MALLOC(byte, XF_Size(fp));
    XF_Read(fpBuf, XF_Size(fp), fp);
    XF_Close(fp);

    fhdr = (BITMAPFILEHEADER*)fpBuf;
	if (fhdr->bfType != 0x4D42) // "BM"
    {
		SYS_FREE(fpBuf);
        return(NULL);
    }
	totalData = fhdr+1;
	info = (BITMAPINFO*)totalData;
	imageData = (byte *)totalData
		+ fhdr->bfOffBits - sizeof(BITMAPFILEHEADER);
	
    result = VIDW_LoadImage((byte*)imageData, info->bmiHeader.biWidth,
        -((int)info->bmiHeader.biHeight), info->bmiHeader.biBitCount, (dword)userData);
    if (result && info->bmiHeader.biBitCount==8)
    {
        for (i=0;i<256;i++)
        {
            VID_SetPaletteEntry(i, info->bmiColors[i].rgbRed, info->bmiColors[i].rgbGreen,
                info->bmiColors[i].rgbBlue);
        }
    }    
    SYS_FREE(fpBuf);
    return((void*)result);
}

static void BmpResourceFree(res_id_t id)
{
    dword imageHandle = (dword)RES_PtrForId(id, NULL);
    VIDW_FreeImage(imageHandle);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================
void VID_Init()
{
    VIDW_Init();
}

void VID_Shutdown()
{
    VIDW_Shutdown();
}

void VID_Frame()
{
    VIDW_Frame();
}

res_id_t VID_RegisterImage(char* fileName, bbool masking)
{
	dword maskDword = masking;
    return(RES_Register(fileName, RES_CACHE_LEVEL, BmpResourceLoad, BmpResourceFree, (void*)maskDword));
}

dword VID_LoadImage(byte* imageData, sdword width, sdword height, dword bitDepth, bbool masking)
{
    return(VIDW_LoadImage(imageData, width, height, bitDepth, masking));
}

void VID_FreeImage(dword imageHandle)
{
    VIDW_FreeImage(imageHandle);
}

void VID_DrawImage(dword imageHandle, sdword scrX, sdword scrY)
{
    VIDW_DrawImage(imageHandle, scrX, scrY);
}

byte* VID_LockImage(dword imageHandle, dword* outImgPitch)
{
    return(VIDW_LockImage(imageHandle, outImgPitch));
}

void VID_UnlockImage(dword imageHandle, byte* lockPtr)
{
    VIDW_UnlockImage(imageHandle, lockPtr);
}

dword VID_GetImageWidth(dword imageHandle)
{
    return(VIDW_GetImageWidth(imageHandle));
}

dword VID_GetImageHeight(dword imageHandle)
{
    return(VIDW_GetImageHeight(imageHandle));
}

dword VID_GetImageBitDepth(dword imageHandle)
{
    return(VIDW_GetImageBitDepth(imageHandle));
}

bbool VID_SetMode(void* mode)
{
    return(VIDW_SetMode(mode));
}

bbool VID_SetModeEx(dword width, dword height, dword bitDepth)
{
    return(VIDW_SetModeEx(width, height, bitDepth));
}

dword VID_GetAvailableModes(void** modeList, dword maxModes)
{
    return(VIDW_GetAvailableModes(modeList, maxModes));
}

dword VID_GetModeWidth(void* mode)
{
    return(VIDW_GetModeWidth(mode));
}

dword VID_GetModeHeight(void* mode)
{
    return(VIDW_GetModeHeight(mode));
}

dword VID_GetModeBitDepth(void* mode)
{
    return(VIDW_GetModeBitDepth(mode));
}

char* VID_GetModeDeviceName(void* mode)
{
    return(VIDW_GetModeDeviceName(mode));
}

char* VID_GetModeDeviceDescription(void* mode)
{
    return(VIDW_GetModeDeviceDescription(mode));
}

void VID_SetPaletteRange(byte* buffer, dword startColor, dword numColors)
{
    VIDW_SetPaletteRange(buffer, startColor, numColors);
}

void VID_SetPaletteEntry(byte color, byte r, byte g, byte b)
{
    VIDW_SetPaletteEntry(color, r, g, b);
}

void VID_GetPaletteEntry(byte color, byte *r, byte *g, byte *b)
{
    VIDW_GetPaletteEntry(color, r, g, b);
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE VID_MAIN.CPP
//**
//****************************************************************************

